/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003 2004  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jet;

import java.awt.Point;
import javax.swing.Icon;
import javax.swing.event.*;

/**
 * Model for palette buttons.
 *
 * @author Adam Domaski
 * @version 0.1 18/06/2004
 */
public abstract class PaletteButtonModel
{
	/**
	 * Gets handlers count in this model.
	 */
	public abstract int getHandlerCount();

	/**
	 * Returns point of handler with specified index. Returned point is in the icon coordinates.
	 *
	 * @param i Index of the handler.
	 * @return Point in icon coordinates.
	 */
	public abstract Point handlerToPoint(int i);

	/**
	 * Returns index of handler with specified point. If there is no handler at this point, it returns -1.
	 *
	 * @param p Point in icon coordiantes.
	 * @return Index of handler or -1 when ther is no handler.
	 */
	public abstract int pointToHandler(Point p);

	/**
	 * Gets an icon of this model.
	 */
	public abstract Icon getIcon();

	/**
	 * Gets tool tip text.
	 */
	public abstract String getToolTipText();

	/**
	 * Sets selected handler by its index.
	 */
	public void setSelectedHandler(int i)
	{
		if (i == selectedHandlerIndex)
			return;
		
		if (i < getHandlerCount() && i >= 0)
		{
			selectedHandlerIndex = i;
			fireStateChanged();
		}
	}

	/**
	 * Get index of selected handler.
	 */
	public int getSelectedHandler()
	{
		return selectedHandlerIndex;
	}

	/**
	 * Clears selection.
	 */
	public void clearSelection()
	{
		selectedHandlerIndex = -1;
		fireStateChanged();
	}

	/**
	 * Returns true if selection is empty.
	 */
	public boolean isSelectionEmpty()
	{
		return (selectedHandlerIndex == -1);
	}

	/**
	 * Registers change listener with this object. Listener will
	 * be notified each time when selection state changes.
	 */
	public void addChangeListener(ChangeListener listener)
	{
		listenerList.add(ChangeListener.class, listener);
	}

	/**
	 * Unregisters change listener.
	 */
	public void removeChangeListener(ChangeListener listener)
	{
		listenerList.remove(ChangeListener.class, listener);
	}

	/**
	 * Fires change event to all registered change listeners.
	 */
	protected void fireStateChanged()
	{
		Object[] listeners = listenerList.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2)
		{
			if (listeners[i] == ChangeListener.class)
			{
				if (changeEvent == null)
					changeEvent = new ChangeEvent(this);
				((ChangeListener) listeners[i+1]).stateChanged(changeEvent);
			}
		}
	}

	/*
	 * --------------- Members ---------------
	 */

	/** Listeners registered with this object. */
	protected EventListenerList listenerList = new EventListenerList();

	/** Selected handle index. */
	private int selectedHandlerIndex = -1;

	/** Dummy event object. */
	private ChangeEvent changeEvent = null;
}
