/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003 2004  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jet;

import java.io.File;
import java.io.FileFilter;
import javax.swing.*;
import java.util.*;

/**
 * Catalog is the class that integrates with existing <b>MaSzyna</b> directory on the
 * filesystem and retrives contents of some subdirectories.
 *
 * @author Adam Domaski
 * @version 0.1 12/03/2004
 */
public class SimulatorCatalog
{
	private static SimulatorCatalog aInstance = null;

	private SimulatorCatalog(Configuration conf)
	{
		createTextures(conf);
	}

	/**
	 * Constructs one instance of this object.
	 *
	 * @param conf Configuration of the <b>jET</b> application.
	 */
	public static void initialize(Configuration conf)
	{
		if (aInstance == null)
			aInstance = new SimulatorCatalog(conf);
	}

	/**
	 * Gets one instance of simulator catalog. It must be initalized before
	 * calling this method.
	 *
	 * @return The instance of the class. It may be null when it wasn't be
	 *   initialized.
	 * @see #initialize
	 */
	public static SimulatorCatalog getInstance()
	{
		return aInstance;
	}

	public MutableComboBoxModel getBmpModel()
	{
		return bmpModel;
	}

	public MutableComboBoxModel getTgaModel()
	{
		return tgaModel;
	}

	/*
	 * ----------- Dirty work ------------
	 */

	private void createTextures(Configuration conf)
	{
		Object o = conf.get("TextureDirectory");
		if (o != null)
		{
			File texDir = new File((String) o);

			bmpModel = new FileListingModel();
			searchBmpTextures(texDir, null);
			Collections.sort(bmpModel.list, fileComparator);			
			bmpModel.list.add(0, "none");

			tgaModel = new FileListingModel();			
			searchTgaTextures(texDir, null);
			Collections.sort(tgaModel.list, fileComparator);
			tgaModel.list.add(0, "none");
		}
	}

	private void searchBmpTextures(File dir, String parentName)
	{
		String prefix;
		String name;
		int last;

		File[] listing = dir.listFiles(bmpFilter);
		if (listing != null)
		{
			int n = listing.length;
			for (int i = 0; i < n; i++)
			{
				prefix = (parentName != null ? parentName + File.separator : "");
				if (listing[i].isDirectory())
					searchBmpTextures(listing[i], prefix + listing[i].getName());
				else
				{
					name = prefix + listing[i].getName();
					last = name.toLowerCase().lastIndexOf(".bmp");					
					bmpModel.list.add( name.substring(0, last) );
				}
			}
		}
	}

	private void searchTgaTextures(File dir, String parentName)
	{
		String prefix;
		File[] listing = dir.listFiles(tgaFilter);
		if (listing != null)
		{
			int n = listing.length;
			for (int i = 0; i < n; i++)
			{
				prefix = (parentName != null ? parentName + File.separator : "");
				if (listing[i].isDirectory())
					searchTgaTextures(listing[i], prefix + listing[i].getName());
				else
					tgaModel.list.add( prefix + listing[i].getName() );
			}
		}
	}

	private FileFilter bmpFilter = new BmpFilter();
	private FileFilter tgaFilter = new TexTgaFilter();

	private FileListingModel bmpModel;
	private FileListingModel tgaModel;

	private Comparator fileComparator = new FilenameComparator();

	/*
	 * ---------- File filters -------------
	 */

	class BmpFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			if (file.isDirectory())
				return true;

			return (file.getName().toLowerCase().endsWith(".bmp"));			
		}
	}

	class TexTgaFilter implements FileFilter
	{
		public boolean accept(File file)
		{
			if (file.isDirectory())
				return true;

			return (file.getName().toLowerCase().endsWith(".tex") ||
				file.getName().toLowerCase().endsWith(".tga"));
		}
	}

	/*
	 * ---------- Combo models --------------
	 */
	
	private class FileListingModel extends AbstractListModel
		implements MutableComboBoxModel
	{
		public List list = new ArrayList();

		public Object getElementAt(int i)
		{
			if (i >= 0 && i < list.size())
				return list.get(i);
			else
				return null;
		}

		public int getSize()
		{
			return list.size();
		}

		public void addElement(Object o)
		{
			list.add(o);
			fireIntervalAdded(this, list.size()-1, list.size()-1);
			if (list.size() == 1 && selectedItem == null && o != null)
			{
				setSelectedItem( o );
			}	
		}

		public void insertElementAt(Object o, int i) {}

		public void removeElement(Object o)	{}

		public void removeElementAt(int i) {}

		public Object getSelectedItem()
		{
			return selectedItem;
		}

		public void setSelectedItem(Object o)
		{
			if ((selectedItem != null && !selectedItem.equals( o )) ||
				selectedItem == null && o != null)
			{
				selectedItem = o;
			    fireContentsChanged(this, -1, -1);
			}
		}

		private Object selectedItem = null;
	}

	/*
	 * ---------- File comparator -----------
	 */
	
	class FilenameComparator implements Comparator
	{
		public int compare(Object o1, Object o2)
		{
			String s1 = (String) o1;
			String s2 = (String) o2;

			int i1 = s1.indexOf(File.separator);
			int i2 = s2.indexOf(File.separator);

			if (i1 == -1 && i2 == -1)
				return s1.compareToIgnoreCase(s2);
			if (i1 == -1)
				return 1;
			if (i2 == -1)
				return -1;

			int result = s1.substring(0, i1).compareToIgnoreCase(s2.substring(0, i2));

			if (result == 0)
				result = compare(s1.substring(i1 + File.separator.length()),
					s2.substring(i2 + File.separator.length()));

			return result;
		}
	}
}
