/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003 2004  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jet.sce;

import javax.swing.event.EventListenerList;

/**
 * Default scenery selection model.
 *
 * @author Adam Domaski
 * @version 0.1 15/02/2004
 */
public class DefaultScenerySelectionModel
	implements ScenerySelectionModel
{
	/*
	 * -------------- Constructors ----------------
	 */
	
	/**
	 * Creates new default scenery selection model for specified scenery 
	 * editor.
	 * The selection is empty at start.
	 */
	public DefaultScenerySelectionModel(JSceneryEditor editor)
	{
		this.editor = editor;
		selectedValue = null;
	}

	/**
	 * Gets scenery editor for this selection model.
	 */
	public JSceneryEditor getEditor()
	{
		return editor;
	}

	/*
	 * ------------- Listeners ----------------
	 */
	
	/**
	 * Binds a new scenery selection listener.
	 *
	 * @see #removeScenerySelectionListener
	 */
	public void addScenerySelectionListener(ScenerySelectionListener l)
	{
		listenerList.add(ScenerySelectionListener.class, l);
	}

	/**
	 * Unbinds a scenery slection listener.
	 *
	 * @see #addScenerySelectionListener
	 */
	public void removeScenerySelectionListener(ScenerySelectionListener l)
	{
		listenerList.remove(ScenerySelectionListener.class, l);
	}

	/**
	 * Gets the scenery selection listeners.
	 */
	public ScenerySelectionListener[] getScenerySelectionListeners()
	{
		return (ScenerySelectionListener[])
			listenerList.getListeners(ScenerySelectionListener.class);
	}

	/**
	 * Notifies scenery selection listeners that selection changed.
	 *
	 * @param event Event to fire.
	 */
	protected void fireSelectionChanged(ScenerySelectionEvent event)
	{
		Object[] listeners = listenerList.getListenerList();
		for (int i = listeners.length - 2; i >= 0; i -= 2)
		{
			if (listeners[i] == ScenerySelectionListener.class)
			{
				 ((ScenerySelectionListener) listeners[i + 1]).valueChanged(event);
			}
		}
	}

	/*
	 * --- Inserting and removing selection ---
	 */

	/**
	 * Clears selection.
	 */
	public void clearSelection()
	{
		selectedValue = null;
		ScenerySelectionEvent event = new ScenerySelectionEvent(this, null);
		fireSelectionChanged(event);
	}

	/**
	 * Returns <code>true</code>, when scenery selection is empty.
	 */
	public boolean isSelectionEmpty()
	{
		return (selectedValue == null);
	}

	/**
	 * Gets current selection.
	 */
	public Object getSelectedObject()
	{
		return selectedValue;
	}

	/**
	 * Selects an object.
	 * The object has to be contained in the scenery.
	 *
	 * @param obj Object to select.
	 */
	public void setSelectedObject(Object obj)
	{
		boolean containing = false;

		// Sprawd czy obiekt jest w scenerii.
		Scenery scenery = editor.getScenery();
		if (scenery != null)
		{
			scenery.readLock();
			try
			{
				containing = scenery.contains(obj);
			}
			finally
			{
				scenery.readUnlock();		
			}
		}

		if (containing)
		{
			selectedValue = obj;
			ScenerySelectionEvent event = new ScenerySelectionEvent(this, obj);
			fireSelectionChanged(event);
		}
	}

	/*
	 * -------------- Members -----------------
	 */
	
	/** Editor for this scenery selection model. */
	private JSceneryEditor editor;

	/** Listener list. */
	protected EventListenerList listenerList = new EventListenerList();

	/** Selected object. */
	private Object selectedValue;
}
