/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003 2004  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package org.jet.tracks;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import org.jet.*;

/**
 * Loads track palette from XML file.
 *
 * @author Adam Domaski
 * @version 0.1 19/06/2004
 */
public class ScriptLoader
{
	/** An instance. */
	private static ScriptLoader anInstance = null;

	/** Track palette. */
	private TrackPalette palette;

	/**
	 * XML file location.
	 */
	public static final String FILENAME = "scripts/tracks.xml";

	private ScriptLoader()
	{
	}

	/**
	 * Returns instance of this class.
	 */
	public static ScriptLoader getInstance()
	{
		if (anInstance == null)
			anInstance = new ScriptLoader();

		return anInstance;
	}

	/**
	 * Loads tracks from XML file.
	 *
	 * @param tp Track palette.
	 * @exception Exception Throws an exception when parsing XML document
	 *   fails.
	 */
	public void load(TrackPalette tp) throws Exception
	{
		palette = tp;

		SAXParserFactory factory = SAXParserFactory.newInstance();
		SAXParser parser = factory.newSAXParser();
		parser.parse(FILENAME, new Loader());
	}

	/**
	 * XML content handler.
	 */
	public class Loader extends DefaultHandler
	{
		/** Mode of first XML element. */
		private int mode = 0;

		private static final int MODEL_STRAIGHT = 1;
		private static final int MODEL_ARC = 2;
		private static final int MODEL_RSWITCH = 3;
		private static final int MODEL_LSWITCH = 4;
		private static final int MODEL_FLEX = 5;

		/** Button model. */
		private PaletteButtonModel buttonModel;

		/* Track or switch. */
		private Track track = null;
		private Switch sswitch = null;

		/** Element */
		private String elementValue = null;

		/* X, Y, Z coordinates. */
		private double x = 0.0;
		private double y = 0.0;
		private double z = 0.0;

		public void startElement(String uri, String lName, String qName, Attributes attrs)
			throws SAXException
		{
			createModel(qName, attrs);
			createTrack(qName);
		}

		private void createModel(String s, Attributes a)
		{
			if (s.equals("straight"))
			{
				double l = Double.parseDouble( a.getValue("length") );
				buttonModel = new StraightTrackButtonModel(l);
				mode = MODEL_STRAIGHT;
			}
			else if (s.equals("arc"))
			{
				double l = Double.parseDouble( a.getValue("length") );
				double r = Double.parseDouble( a.getValue("radius") );
				buttonModel = new ArcTrackButtonModel(l, r);
				mode = MODEL_ARC;
			}
			else if (s.equals("sswitch"))
			{
				String side = a.getValue("side");
				double l = Double.parseDouble( a.getValue("length") );
				double r = Double.parseDouble( a.getValue("radius") );
				String sk = a.getValue("skos");
				if (side.equals("left"))
				{
					buttonModel = new LeftSwitchButtonModel(l, r, sk);
					mode = MODEL_LSWITCH;
				}
				else if (side.equals("right"))
				{
					buttonModel = new RightSwitchButtonModel(l, r, sk);
					mode = MODEL_RSWITCH;
				}
			}
			else if (s.equals("flex"))
			{
				String desc = a.getValue("description");
				buttonModel = new FlexTrackButtonModel(desc);
				mode = MODEL_FLEX;
			}
		}

		private void createTrack(String s)
		{
			if (s.equals("track"))
			{
				track = new Track();
			}
			else if (s.equals("switch"))
			{
				sswitch = new Switch();
			}
		}

		public void endElement(String uri, String lName, String qName)
			throws SAXException
		{
			processButtonModel(qName);
			processEntity(qName);
		}

		private void processButtonModel(String s)
		{
			if (s.equals("straight") || s.equals("arc") || s.equals("flex"))
			{
				JPaletteButton jButton = new JPaletteButton(buttonModel, track);
				palette.layout(jButton);
				mode = 0;
				track = null;
			}
			else if (s.equals("sswitch"))
			{
				JPaletteButton jButton = new JPaletteButton(buttonModel, sswitch);
				palette.layout(jButton);
				mode = 0;
				sswitch = null;
			}
		}

		private void processEntity(String s)
		{
			// Rolls
			if (s.equals("roll1"))
			{
				if (track != null)
					track.setRoll1( Double.parseDouble( elementValue ) );
				else
					sswitch.setRoll1( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("roll2"))
			{
				if (track != null)
					track.setRoll2( Double.parseDouble( elementValue ) );
				else
					sswitch.setRoll2( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("roll3"))
			{
				sswitch.setRoll3( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("rolls"))
			{
				sswitch.setSwitchedRoll( Double.parseDouble( elementValue ) );
			}
			// Radiuses
			else if (s.equals("radius"))
			{
				track.setRadius( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("radius1"))
			{
				sswitch.setRadius( Double.parseDouble(elementValue) );
			}
			else if (s.equals("radius2"))
			{
				sswitch.setSwitchedRadius( Double.parseDouble(elementValue) );
			}
			// Length, width...
			else if (s.equals("length"))
			{
				if (track != null)
					track.setLength( Double.parseDouble( elementValue ) );
				else
					sswitch.setLength( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("width"))
			{
				if (track != null)
					track.setWidth( Double.parseDouble( elementValue ) );
				else
					sswitch.setWidth( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("friction"))
			{
				if (track != null)
					track.setFriction( Double.parseDouble( elementValue ) );
				else
					sswitch.setFriction( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("tap"))
			{
				if (track != null)
					track.setTap( Double.parseDouble( elementValue ) );
				else
					sswitch.setTap( Double.parseDouble( elementValue ) );
			}
			else if (s.equals("quality"))
			{
				if (track != null)
					track.setQuality( Integer.parseInt( elementValue ) );
				else
					sswitch.setQuality( Integer.parseInt( elementValue ) );
			}
			else if (s.equals("damage"))
			{
				if (track != null)
					track.setDamage( Integer.parseInt( elementValue ) );
				else
					sswitch.setDamage( Integer.parseInt( elementValue ) );
			}
			else if (s.equals("enviroment"))
			{
				if (track != null)
					track.setEnviroment( elementValue );
				else
					sswitch.setEnviroment( elementValue );
			}
			else if (s.equals("visibility"))
			{
				if (track != null)
					track.setVisible( Boolean.valueOf( elementValue ).booleanValue() );
				else
					sswitch.setVisible( Boolean.valueOf( elementValue ).booleanValue() );
			}
			// Textures
			else if (s.equals("texture"))
			{
				if (track != null)
					track.setTexture( elementValue );
				else
					sswitch.setTexture( elementValue );
			}
			else if (s.equals("texlength"))
			{
				if (track != null)
					track.setTextureLength( Double.parseDouble(elementValue) );
				else
					sswitch.setTextureLength( Double.parseDouble(elementValue) );
			}
			else if (s.equals("secondtexture"))
			{
				if (track != null)
					track.setSecondTexture( elementValue );
				else
					sswitch.setSecondTexture( elementValue );
			}
			else if (s.equals("texwidth"))
			{
				if (track != null)
					track.setTextureWidth( Double.parseDouble(elementValue) );
				else
					sswitch.setTextureWidth( Double.parseDouble(elementValue) );
			}
			else if (s.equals("texheight"))
			{
				if (track != null)
					track.setTextureHeight( Double.parseDouble(elementValue) );
				else
					sswitch.setTextureHeight( Double.parseDouble(elementValue) );
			}
			else if (s.equals("texslope"))
			{
				if (track != null)
					track.setTextureSlope( Double.parseDouble(elementValue) );
				else
					sswitch.setTextureSlope( Double.parseDouble(elementValue) );
			}
			// X, Y, Z coordinates
			else if (s.equals("x"))
			{
				x = Double.parseDouble(elementValue);
			}
			else if (s.equals("y"))
			{
				y = Double.parseDouble(elementValue);
			}
			else if (s.equals("z"))
			{
				z = Double.parseDouble(elementValue);
			}
			// Coordinates
			else if (s.equals("node1"))
			{
				if (track != null)
					track.setPoint1(x, y, z);
				else
					sswitch.setPoint1(x, y, z);
			}
			else if (s.equals("node2"))
			{
				if (track != null)
					track.setPoint2(x, y, z);
				else
					sswitch.setPoint2(x, y, z);
			}
			else if (s.equals("vector1"))
			{
				if (track != null)
					track.setVector1(x, y, z);
				else
					sswitch.setVector1(x, y, z);
			}
			else if (s.equals("vector2"))
			{
				if (track != null)
					track.setVector2(x, y, z);
				else
					sswitch.setVector2(x, y, z);
			}
			else if (s.equals("node3"))
			{
				sswitch.setPoint3(x, y, z);
			}
			else if (s.equals("vector3"))
			{
				sswitch.setVector3(x, y, z);
			}
			else if (s.equals("vectors"))
			{
				sswitch.setSwitchedVector(x, y, z);
			}
		}

		public void characters(char[] ch, int start, int len)
			throws SAXException
		{
			elementValue = new String(ch, start, len);
		}
	}
}
