/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package prv.doman.fsa;

import java.io.Serializable;
import java.util.*;

/**
 * The automaton is a short for FSA (finit state automaton).
 *
 * @author Adam Domaski
 * @version 0.1 21/08/2003
 */
public class Automaton implements Serializable
{
	/**
	 * Creates the new finit state automaton.
	 */
	public Automaton()
	{
		alphabet = new HashSet();
		states = new HashSet();
		finalStates = new HashSet();
		rules = new HashMap();
	}

	/*
	 * ------------- Setting up automaton ----------------
	 */

	/**
	 * Adds a letter to the alphabet.
	 *
	 * @param letter New letter to add.
	 */
	public void addLetter(String letter)
	{
		alphabet.add(letter);
	}

	/**
	 * Adds a new state for automaton.
	 *
	 * @param stateName Name of the new state.
	 * @exception StateRedefinitionException Thrown when adding state
	 *   already exists in the set of states.
	 */
	public void addState(String stateName) throws StateRedefinitionException
	{
		if (states.contains(stateName))
			throw new StateRedefinitionException(stateName);

		states.add(stateName);
	}

	/**
	 * Adds a transition between states as result of occuring some letter
	 * from alphabet at input.
	 *
	 * @param letter Input.
	 * @param startState Start state name.
	 * @param endState End state name determined by input and start name.
	 * @exception UnknownLetterException Thrown when letter is not in
	 *   automaton's alphabet.
	 * @exception UnknownStateException Thrown when start state or end state
	 *   are not in set of states.
	 */
	public void addRule(String letter, String startState, String endState)
		throws UnknownLetterException, UnknownStateException
	{
		if (!alphabet.contains(letter))
			throw new UnknownLetterException(letter);

		if (!states.contains(startState))
			throw new UnknownStateException(startState);

		if (!states.contains(endState))
			throw new UnknownStateException(endState);

		rules.put(new Rule(startState, letter), endState);
	}

	/**
	 * Adds new rule that is default for transitions between start state
	 * and end state.
	 *
	 * @param startState Start state name.
	 * @param endState End state name determined by input and start name.
	 * @exception UnknownStateException Thrown when start state or end state
	 *   are not in set of states.
	 */
	public void addDefaultRule(String startState, String endState)
		throws UnknownStateException
	{
		if (!states.contains(startState))
			throw new UnknownStateException(startState);

		if (!states.contains(endState))
			throw new UnknownStateException(endState);

		Iterator i = alphabet.iterator();
		while (i.hasNext())
		{
			String letter = (String) i.next();
			Rule rule = new Rule(startState, letter);
			if ( !rules.containsKey(rule) )
				rules.put(rule, endState);
		}
	}

	/**
	 * Sets state as final state.
	 * <p>State must be earlier added to set of states.</p>
	 *
	 * @param stateName Name of state to be final.
	 * @exception UnknownStateException Thrown when set of states
	 *   does not include this state.
	 */
	public void setFinal(String stateName) throws UnknownStateException
	{
		if (!states.contains(stateName))
			throw new UnknownStateException(stateName);

		finalStates.add(stateName);
	}

	/**
	 * Sets start state.
	 * <p>State must have been added to set of states.</p>
	 *
	 * @param stateName Name of state.
	 * @exception UnknownStateException Thrown when set of states
	 *   does not include this start state.
	 */
	public void setStart(String stateName) throws UnknownStateException
	{
		if (!states.contains(stateName))
			throw new UnknownStateException(stateName);

		state = stateName;
	}

	/*
	 * ---------- Looping -----------
	 */
	
	/**
	 * Makes a step.
	 *
	 * @param letter Input.
	 * @exception UnknownLetterException Thrown when letter is not from
	 *   alphabet.
	 * @exception FinalStateMissing Thrown when automaton does not have at least
	 *   one final state.
	 */
	public void step(String letter)
		throws UnknownLetterException, FinalStateMissingException
	{
		if (!alphabet.contains(letter))
			throw new UnknownLetterException(letter);

		if (finalStates.size() == 0)
			throw new FinalStateMissingException();

		state = (String) rules.get(new Rule(state, letter));
	}

	/*
	 * --------- Data retriving ---------
	 */

	/**
	 * Retrives current state name.
	 *
	 * @return State name.
	 */
	public String currentState()
	{
		return state;
	}

	/**
	 * Checks if <code>letter</code> is in alphabet.
	 *
	 * @param letter Letter to check.
	 * @return <code>True</code>, if alphabet contains this letter.
	 */
	public boolean hasLetter(String letter)
	{
		return alphabet.contains(letter);
	}

	/**
	 * Gets all state names.
	 *
	 * @return Array of all state names.
	 */
	public String[] stateNames()
	{
		return (String[]) states.toArray();
	}

	/**
	 * Checks if current state is final state.
	 *
	 * @return <code>True</code>, if current state is final.
	 */
	public boolean isFinished()
	{
		return finalStates.contains(state);
	}
	
	/**
	 * Gets alphabet.
	 *
	 * @return Array of letter of alphabet.
	 */
	public String[] getAlphabet()
	{
		return (String[]) alphabet.toArray();
	}

	/*
	 * ------------ Inner class --------------
	 */
	
	/**
	 * This is a rule for finite state automaton.
	 * The rule is a pair of initial state and letter.
	 */
	public class Rule
	{
		/**
		 * Creates new rule object.
		 *
		 * @param state State name.
		 * @param letter The letter.
		 */
		public Rule(String state, String letter)
		{
			this.state = state;
			this.letter = letter;
		}

		/**
		 * Gets state name for this rule.
		 */
		public String getState()
		{
			return state;
		}

		/**
		 * Sets state name for this rule.
		 */
		public void setState(String state)
		{
			this.state = state;
		}

		/**
		 * Gets a letter in this rule.
		 */
		public String getLetter()
		{
			return letter;
		}

		/**
		 * Sets the letter in this rule.
		 */
		public void setLetter(String letter)
		{
			this.letter = letter;
		}

		/**
		 * Returns true if objects are equal.
		 */
		public boolean equals(Object o)
		{
			Rule rule = (Rule) o;
			if (state.equals(rule.getState()) && 
				letter.equals(rule.getLetter()))
			{
				return true;
			}

			return false;
		}

		/**
		 * Computes new hash code used in hastables.
		 * It just multiplies state name code with letter hash code.
		 */
		public int hashCode()
		{
			return state.hashCode() * letter.hashCode();
		}

		/**
		 * Gets string description of this object.
		 */
		public String toString()
		{
			StringBuffer sb = new StringBuffer();
			sb.append("Rule[");
			sb.append(state);
			sb.append(',');
			sb.append(letter);
			sb.append(']');
			return sb.toString();
		}

		/*
		 * ----------- Members -------------
		 */
		private String state;
		private String letter;
	}

	/*
	 * -------------- Members ----------------
	 */
	
	/** Current state of automaton */
	private String state;

	/** Alphabet. */
	private Set alphabet;

	/** States. */
	private Set states;

	/** Final states. */
	private Set finalStates;

	/** Rules. */
	private Map rules;
}
