/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003 2004  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package prv.doman.util;

import java.text.*;
import java.util.Locale;

/**
 * Formats and parse angle with specified pattern.
 *
 * <p>It formats integer and floating-point numbers. The full output
 * may look like these:</p>
 *
 * <blockquote>
 *   23&#xB0;35&#x2032;03.234&#x2033;<br>
 *   1,567&#xB0;0&#x2032;59&#x2033;<br>
 *   0&#xB0;00.500&#x2032;<br>
 *   -24.75&#xB0;<br>
 *   90&#xB0;
 * </blockquote>
 *
 * <p>So output seems to have at most four fields: degrees, minutes, seconds and 
 * the fractional part of seconds. Note that symbols <code>&#x2032;</code> and <code>&#x2033;</code>
 * aren't the ASCII Quotation Mark (Unicode <code>&#x5C;u0022</code>) and Apostrophe (Unicode 
 * <code>&#x5C;u0027</code>), but Mathematical symbols: Prime (Unicode <code>&#x5C;u2033</code>)
 * and Double Prime (Unicode <code>&#x5C;u2032</code>). Note that decimal separators and fraction dot
 * might be localized.
 *
 * <p>The right order in format must be as follows: degrees, minutes, seconds. The fractional part
 * is always at the and of output and it is optional. Optional are minutes and seconds also, but it
 * is not possible to have format for examaple: degrees and seconds without minutes.</p>
 *
 * <h4>Patterns</h4>
 *
 * <p>The parttern is a string containing characters with special meanings:</p>
 *
 * <blockquote>
 *   <table cellpadding="3">
 *     <tr><td>D</td><td>The integer part of degree field.</td></tr>
 *     <tr class="TableSubHeadingColor"><td>d</td><td>The fractional part of degree field.</td></tr>
 *     <tr><td>M</td><td>The integer part of minute field.</td></tr>
 *     <tr class="TableSubHeadingColor"><td>m</td><td>The fractional part of minute field.</td></tr>
 *     <tr><td>S</td><td>The integer part of second fields.</td></tr>
 *     <tr class="TableSubHeadingColor"><td>s</td><td>The fractional part of second field.</td></tr>
 *     <tr><td>.</td><td>Fractional separator (localized).</td></tr>
 *     <tr class="TableSubHeadingColor"><td>&#xB0; (Unicode &#x5C;u00B0)</td><td>Degree symbol.</td></tr>
 *     <tr><td>&#x2032; (Unicode &#x5C;u2032)</td><td>Minute symbol</td></tr>
 *     <tr class="TableSubHeadingColor"><td>&#x2033; (Unicode &#x5C;u2033)</td><td>Second symbol.</td></tr>
 *   </table>
 * </blockquote>
 *
 * <p>Here are some expamples of patterns:</p>
 *
 * <blockquote>
 *   <table cellpadding="3">
 *     <tr class="TableHeadingColor" align="center">
 *        <td><b>Pattern</b></td><td colspan="2"><b>Examples</b></td>
 *     </tr>
 *     <tr><td>D&#xB0;</td><td>360&#xB0;</td><td>2&#xB0;</td></tr>
 *     <tr class="TableSubHeadingColor"><td>DDD.dd&#xB0;</td><td>008.55&#xB0;</td><td>241.08&#xB0;</td>
 *       </tr>
 *     <tr><td>D&#xB0;M.m&#x2032;</td><td>123&#xB0;20.2&#x2032;</td><td>-3&#xB0;2.0&#x2032;</td></tr>
 *     <tr class="TableSubHeadingColor"><td>DDDMMSS</td><td>1803000</td><td>0030000</td></tr>
 *   </table>
 * </blockquote>
 *
 *
 * <p>For symbol customization use class {@link DegreeFormatSymbols DegreeFormatSymbols}.</p>
 * 
 * @author Adam Domaski
 * @version 0.1 16/02/2004
 */
public class DegreeFormat extends Format 
{
	/**
	 * The degree field.
	 */
	public static final int DEGREE_FIELD = 0;

	/**
	 * The minute field.
	 */
	public static final int MINUTE_FIELD = 1;

	/**
	 * The second field.
	 */
	public static final int SECOND_FIELD = 2;
	
	// ------------------ Constructors -----------------

	/**
	 * Constructs degree format with default pattern and symbols.
	 * Default pattern is <code>D&#xB0;MM&#x2032;SS&#x2033;</code>.
	 */
	public DegreeFormat()
	{
		symbols = new DegreeFormatSymbols();
		applyPattern("D\u00B0MM\u2032SS\u2033");

		DecimalFormatSymbols dfs = decimalFormat.getDecimalFormatSymbols();
		dfs.setDecimalSeparator( symbols.getDecimalSeparator() );
	}

	/**
	 * Constructs degree format with specified pattern.
	 * Symbols are default.
	 *
	 * @param pattern Pattern string.
	 */
	public DegreeFormat(String pattern)
	{
		symbols = new DegreeFormatSymbols();
		applyPattern(pattern);

		DecimalFormatSymbols dfs = decimalFormat.getDecimalFormatSymbols();
		dfs.setDecimalSeparator( symbols.getDecimalSeparator() );
	}

	/**
	 * Constructs degree format with specified pattern and symbols.
	 *
	 * @param pattern Pattern string.
	 * @param symbols Custom symbols.
	 */
	public DegreeFormat(String pattern, DegreeFormatSymbols symbols)
	{
		this.symbols = symbols;
		applyPattern(pattern);

		DecimalFormatSymbols dfs = decimalFormat.getDecimalFormatSymbols();
		dfs.setDecimalSeparator( symbols.getDecimalSeparator() );
	}

	// -------------- Symbols -------------

	/**
	 * Sets for this format symbols.
	 *
	 * @param s Degree format symbols.
	 */
	public void setDegreeFormatSymbols(DegreeFormatSymbols s)
	{
		symbols = s;
	}

	/**
	 * Gets the format symbols of this format.
	 *
	 * @return Degree format symbols.
	 */
	public DegreeFormatSymbols getDegreeFormatSymbols()
	{
		return symbols;
	}

	/**
	 * Gets a symbol at given field position.
	 *
	 * @param field Identifier of degree fields.
	 * @return <code>&#x00B0;</code> or <code>&#x2032;</code> or <code>&#x2033;</code>.
	 */
	private char getSymbol(int field)
	{
		switch (field)
		{
		case DEGREE_FIELD:
			return symbols.getDegree();
		case MINUTE_FIELD:
			return symbols.getMinute();
		case SECOND_FIELD:
			return symbols.getSecond();
		}

		return symbols.getDegree();
	}

	// -------------- Patterns ----------------

	private int[] widths = {0, 0, 0};
	static private char[] charts = {'D', 'M', 'S'};
	private boolean[] hasSymbol = {false, false, false};
	private int fracWidth = 0;

	/**
	 * Applies pattern to this format.
	 * Check the pattern if it has correct syntax.	 
	 *
	 * @param pattern A pattern.
	 * @exception IllegalArgumentException When pattern has illegal syntax.
	 */
	public void applyPattern(String pattern)
	{
		boolean hasDot = false;
		int stage = DEGREE_FIELD;

		for (int i = 0; i < pattern.length(); i++)
		{
			char ch = pattern.charAt(i);

			if (stage < SECOND_FIELD)
			{
				if (ch == charts[ stage+1 ] && widths[stage] > 0)
				{
					stage++;
				}
			}

			if (ch == charts[ stage ] && !hasDot)
			{
				widths[ stage ]++;
				if (stage > DEGREE_FIELD && widths[ stage ] > 2)
				{
					throw new IllegalArgumentException("Field "+stage+" has incorrect syntax");
				}
			}
			else if (ch == Character.toLowerCase( charts[ stage ] ))
			{
				fracWidth++;
			}
			else if (ch == '.' && !hasDot && fracWidth == 0)
			{
				hasDot = true;
			}
			else if (ch == '\u00b0' || ch == '\u2032' || ch == '\u2033')
			{
				hasSymbol[ stage ] = true;
			}
			else
			{
				throw new IllegalArgumentException("Pattern has incorrect syntax");
			}
		}
	}

	/**
	 * Gets pattern of this format.
	 *
	 * @return A pattern.
	 */
	public String toPattern()
	{
		return pattern;
	}

	// -------------- Locales -----------------

	/**
	 * Gets the instance of degree format class with default locale.
	 * The pattern is default: <code>D&#xB0;MM&#x2033;SS&#x2032;</code>.
	 */
	public static final DegreeFormat getInstance()
	{
		return new DegreeFormat();
	}

	/**
	 * Gets the instace of degree format class with specified locale.
	 * The pattern is default: <code>D&#xB0;MM&#x2033;SS&#x2032;</code>.
	 *
	 * @param locale lo
	 */
	public static final DegreeFormat getInstance(Locale locale)
	{
		DegreeFormat df = new DegreeFormat();
		df.setDegreeFormatSymbols(new DegreeFormatSymbols(locale));
		return df;
	}

	/**
	 * Returns all available locales in Java Virtual Machine.
	 *
	 * @return Array of locales.
	 */
	public static Locale[] getAvailableLocales()
	{
		return Locale.getAvailableLocales();
	}

	// ------------------ Formats ----------------------

	/**
	 * Tries to format object with output redirected to string buffer and with
	 * position.
	 * Object must be a number class.
	 *
	 * @param obj A number to format.
	 * @param toAppendTo String buffer where append to formatted output.
	 * @param pos Field position (unused).
	 * @return String buffer with appended formatted output.
	 * @exception IllegalArgumentException When object is not a number.
	 */
	public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition pos)
	{
		if (obj instanceof Number)
		{
			return format(((Number) obj).doubleValue(), toAppendTo, pos);
		}
		else
		{
			throw new IllegalArgumentException("Cannot format given Object as a Number");
		}
	}

	/**
	 * Formats simple double value.
	 *
	 * @param number Floating point number to format.
	 * @return String containing formatted output.
	 */
	public String format(double number)
	{
		return format(number, new StringBuffer(), new FieldPosition(0)).toString();
	}

	/**
	 * Formats integer value.
	 *
	 * @param number Integer number to format.
	 * @return String containing formatted output.
	 */
	public String format(int number)
	{
		return format((double) number, new StringBuffer(), new FieldPosition(0)).toString();
	}

	/**
	 * Formats floating point number using format's pattern and symbols.
	 * The output will be redirected to string buffer and format starts
	 * a the field position.
	 *
	 * @param number Floating point number going to be formatted.
	 * @param toAppendTo String buffer where to append.
	 * @param pos Field position (unused).
	 * @return String buffer containing output.
	 * @exception NullPointerException When position or string buffer is null.
	 */
	public StringBuffer format(double number, StringBuffer toAppendTo, FieldPosition pos)
	{
		if (toAppendTo == null)
			throw new NullPointerException("String buffer is null");
		if (pos == null)
			throw new NullPointerException("Position is null");

		double degrees = number;
		double minutes = Double.NaN;
		double seconds = Double.NaN;

		boolean addMinus = (degrees > -1.0 && degrees < 0.0);

		if (widths[1] > 0)
		{
			int tmp = (int) degrees;
			minutes = Math.abs(degrees - tmp) * 60;
			degrees = tmp;

			if (widths[2] > 0)
			{
				tmp = (int) minutes;
				seconds = (minutes - tmp) * 60;
				minutes = tmp;

                seconds = Math.rint(seconds);
                tmp = (int) (seconds / 60);
                seconds -= 60 * tmp;
                minutes += tmp;
			}
            else
            {
				minutes = Math.rint(minutes);
            }
            tmp = (int) (minutes / 60);
            minutes -= 60*tmp;
            degrees += tmp;
		}

		if (addMinus)
			degrees = -degrees;

		// Postion advance
		int field = pos.getField();
		pos.setBeginIndex(0);
		pos.setEndIndex(0);

		toAppendTo = formatField(degrees, toAppendTo, (field == DEGREE_FIELD ? pos: null),
			widths[0], widths[1] == 0, DEGREE_FIELD);
		if (!Double.isNaN(minutes))
			toAppendTo = formatField(minutes, toAppendTo, (field == MINUTE_FIELD ? pos : null),
				widths[1], widths[2] == 0, MINUTE_FIELD);
		if (!Double.isNaN(seconds))
			toAppendTo = formatField(seconds, toAppendTo, (field == SECOND_FIELD ? pos : null),
				widths[2], true, SECOND_FIELD);

		return toAppendTo;
	}

	/**
	 * Formats a single field within degree format.
	 *
	 * @param number Number to format (degrees or minutes or seconds).
	 * @param toAppendTo Text buffer where append to.
	 * @param pos Field position.
	 * @param width Width in digits of integer part of number output.
	 * @param last Flag indicating that the field is last in a sequence.
	 * @param filed Field identifier.
	 * @return Formatted output for a single field.
	 */
	private StringBuffer formatField(double number, StringBuffer toAppendTo, FieldPosition pos,
		int width, boolean last, int field)
	{
		int startPosition = toAppendTo.length();

		if (!last)
		{
			decimalFormat.setMinimumIntegerDigits(width);
			decimalFormat.setMaximumFractionDigits(0);
			toAppendTo = decimalFormat.format(number, toAppendTo, dummy);
		}
		else if (fracWidth > 0)
		{
            decimalFormat.setMinimumIntegerDigits(width);
            decimalFormat.setMinimumFractionDigits(fracWidth);
            decimalFormat.setMaximumFractionDigits(fracWidth);
            toAppendTo = decimalFormat.format(number, toAppendTo, dummy);
        }
		else
		{
			decimalFormat.setMinimumIntegerDigits(width);
			decimalFormat.setMaximumFractionDigits(0);
			toAppendTo = decimalFormat.format(number, toAppendTo, dummy);
		}

		if (hasSymbol[field])
		{
			toAppendTo.append(getSymbol(field));
		}

		if (pos != null)
		{
			pos.setBeginIndex(startPosition);
			pos.setEndIndex(toAppendTo.length() - 1);
		}

		return toAppendTo;
	}

	// ---------------- Parsing --------------------

	/**
	 * Parses string to find an object that representing value.
	 *
	 * @param source String to parse.
	 * @param parsePosition Data structure of success or fail in parsing.
	 * @return An object (a double number).
	 */
	public Object parseObject(String source, ParsePosition parsePosition)
	{
		return parse(source, parsePosition);
	}

	/**
	 * Parses string to find the double number representing.
	 *
	 * <p>This method parses numbers:</p>
	 * <ul>
	 *   <li>separated by white space characters,
	 *   <li>separated by Unicode symbols: &#xB0;, &#x2032;, &#x2033;,
	 *   <li>separated by apostrophe and quote symbols ", ';
	 *   <li>that have fields written in following order:
	 *     <blockquote>
	 *        DEGREES [ MINUTES [ SECONDS ] ]
	 *     </blockquote>
	 * </ul>
	 *
	 * @param source String to parse.
	 * @param parsePosition Data structure to inviligate parsing progress.
	 * @return A double number.
	 */
	public Double parse(String source, ParsePosition parsePosition)
	{
		if (parsePosition == null)
			throw new NullPointerException("Parse position is null");

		double degrees = Double.NaN;
		double minutes = Double.NaN;
		double parsedValue = Double.NaN;

		int length = source.length();
		int startPosition = parsePosition.getIndex();
		int index = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Rozbierz pierwsz liczb
		Number fieldObject = decimalFormat.parse(source, parsePosition);
		if (fieldObject == null)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		parsedValue = fieldObject.doubleValue();

		// Sprawd czy wystpia kropka
		boolean dot = hasDot( source.substring(index, parsePosition.getIndex()) );
		if (dot)
		{
			return new Double( fieldObject.doubleValue() );
		}
		index = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Sprawd czy jest symbol stopnia
		if (index < length)
		{
			char ch = source.charAt(index);
			if (ch == '\u00B0' || ch == symbols.getDegree())				
			{
				index++;
			}

			parsePosition.setIndex(index);

			Double minuteField = parseMinutes(source, parsePosition);
			if (minuteField == null)
			{
				parsePosition.setIndex(index);
			}
			else
			{
				minutes = minuteField.doubleValue();
			}
		}

		degrees = parsedValue;

		// Z warto liczby

		if (degrees < 0)
		{
			minutes = -minutes;
		}

		if (Double.isNaN(degrees)) degrees = 0;
		if (!Double.isNaN(minutes)) degrees += minutes/60; 

		return new Double(degrees);
	}

	/**
	 * Parses the minutes field.
	 */
	private Double parseMinutes(String source, ParsePosition parsePosition)
	{
		double minutes = Double.NaN;
		double seconds = Double.NaN;
		double parsedValue;

		int length = source.length();
		int startPosition = parsePosition.getIndex();
		int index = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Rozbierz liczb
		Number fieldObject = decimalFormat.parse(source, parsePosition);
		if (fieldObject == null)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}

		// Sprawd czy warto jest poprawna dla minut
		parsedValue = fieldObject.doubleValue();
		if (!isMinute(parsedValue))
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}

		// Sprawd czy wystpia kropka
		boolean dot = hasDot( source.substring(index, parsePosition.getIndex()) );
		if (dot)
		{
			return new Double(parsedValue);
		}
		index = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Sprawd czy s na kocu symbole minut
		if (index < length)
		{
			char ch = source.charAt(index);
			if (ch == '\"' || ch == '\u2033' || ch == symbols.getMinute())
			{
				index++;
			}

			parsePosition.setIndex(index);

			// Rozbierz pole sekund
			Double secondField = parseSeconds(source, parsePosition);
			if (secondField == null)
			{
				parsePosition.setIndex(index);
			}
			else
			{
				seconds = secondField.doubleValue();
			}
		}

		minutes = parsedValue;

		// Przygotuj warto liczby
		if (minutes < 0)
		{
			seconds = -seconds;
		}

		if (Double.isNaN(minutes)) minutes = 0.0;
		if (!Double.isNaN(seconds)) minutes += seconds / 60;

		return new Double(minutes);
	}

	/**
	 * Parses the seconds field.
	 */
	private Double parseSeconds(String source, ParsePosition parsePosition)
	{
		double seconds = Double.NaN;
		double parsedValue;

		int length = source.length();
		int index = parsePosition.getIndex();
		int startPosition = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Rozbierz liczb
		Number fieldObject = decimalFormat.parse(source, parsePosition);
		if (fieldObject == null)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		
		// Sprawd czy liczba jest z dozwolonego zakresu
		parsedValue = fieldObject.doubleValue();
		if (!isMinute( parsedValue ))
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		index = parsePosition.getIndex();

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Sprawd czy s na kocu symbole sekund
		if (index < length)
		{
			char ch = source.charAt(index);
			if (ch == '\"' || ch == '\u2033' || ch == symbols.getSecond())
			{
				index++;
				parsePosition.setIndex(index);
			}
		}

		seconds = parsedValue;

		// Przejd przez pozostae biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		return new Double(seconds);
	}

	/**
	 * Checks if string has decimal separator inside.
	 *
	 * @param str String to search to.
	 * @return <code>True</code> if string contains decimal separator.
	 */
	private boolean hasDot(String str)
	{
		char dot = symbols.getDecimalSeparator();

		for (int i = 0; i < str.length(); i++)
		{
			char ch = str.charAt(i);
			if (ch == dot)
				return true;
		}

		return false;
	}

	/**
	 * Checks if an absolute value of double is in range [0, 60].
	 *
	 * @param d Double value to check.
	 * @return <code>True</code>, if double is in that range.
	 */
	private boolean isMinute(double d)
	{
		d = Math.abs(d);
		if (d >= 0.0 && d < 60.0)
			return true;

		return false;
	}

	// ------------- Privates --------------

	private DegreeFormatSymbols symbols;
	private String pattern;
	private FieldPosition dummy = new FieldPosition(0);
	private DecimalFormat decimalFormat = new DecimalFormat();
}
