/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package prv.doman.util;

import java.text.Format;
import java.text.FieldPosition;
import java.text.ParsePosition;
import java.text.DecimalFormat;

/**
 * Number format class for scales.
 * <p>Typical scale notation is 1:<i>x</i> or <i>x</i>:1. Numbers must be positive.</p>
 *
 * @author Adam Domaski
 * @version 0.1 19/08/2003
 */
public class ScaleFormat extends Format
{
	/**
	 * Creates scale format.
	 */
	public ScaleFormat()
	{
		decimalFormat = new DecimalFormat();
	}

	/**
	 * Gets the delimeter.
	 *
	 * @return Delimeter.
	 */
	public String getDelimeter()
	{
		return delimeter;
	}

	/**
	 * Sets the delimeter.
	 * A default delimeter is <code>:</code>.
	 *
	 * @param newDelimeter The new delimeter.
	 */
	public void setDelimeter(String newDelimeter)
	{
		delimeter = newDelimeter;
	}

	/*
	 * ------------- Locales --------------
	 */
	
	/**
	 * Gets instance of this class.
	 */
	public static ScaleFormat getInstance()
	{
		return new ScaleFormat();
	}

	/*
	 * ----------- Formatting -------------
	 */

	/**
	 * Formats double value.
	 * Note that value must be positive. Otherwise exception may be thrown.
	 *
	 * @param number Float number.
	 * @return String output.
	 * @exception IllegalArgumentException Thrown when the number isn't positive.
	 */
	public String format(double number)
	{
		return format(number, new StringBuffer(), new FieldPosition(0)).toString();
	}

	/**
	 * Formats integer value.
	 * Note that value must be positive. Otherwise exception may be thrown.
	 *
	 * @param number Integer number to format.
	 * @return String output.
	 * @exception IllegalArgumentException Thrown when the number isn't positive.
	 */
	public String format(int number)
	{
		return format((double) number, new StringBuffer(), new FieldPosition(0)).toString();
	}

	/**
	 * Formats object supposed to be a number.
	 *
	 * @param obj Object to format.
	 * @param toAppendTo String buffer where append output.
	 * @param pos Field position (unused).
	 * @return String output.
	 * @exception IllegalArgumentException Thrown when the object isn't a number nor
	 *   the number is positive.
	 */
	public StringBuffer format(Object obj, StringBuffer toAppendTo, FieldPosition pos)
	{
		if (obj instanceof Number)
		{
			return format(((Number) obj).doubleValue(), toAppendTo, pos);
		}
		else
		{
			throw new IllegalArgumentException("Cannot format given Object as a Number");
		}
	}

	/**
	 * Formats double value and appends it to string output.
	 *
	 * @param number A number to format. Must be positive.
	 * @param toAppendTo The string output to append.
	 * @param pos Field position (unused).
	 * @return String output appended to string buffer.
	 * @exception IllegalArgumentException Thrown when number isn't positive.
	 */
	public StringBuffer format(double number, StringBuffer toAppendTo, FieldPosition pos)
	{
		if (number <= 0.0)
			throw new IllegalArgumentException("Cannot format non-positive numbers");

		if (number >= 1.0)
		{
			toAppendTo.append( decimalFormat.format(number) );
			toAppendTo.append( delimeter );
			toAppendTo.append( '1' );
		}
		else
		{
			number = 1.0 / number;
			toAppendTo.append( '1' );
			toAppendTo.append( delimeter );
			toAppendTo.append( decimalFormat.format(number) );
		}

		return toAppendTo;
	}

	/*
	 * ---------------- Parsing -----------------
	 */
	
	/**
	 * Parses source to find object representing scale.
	 *
	 * @param source String to parse.
	 * @param parsePosition Parse position which tracks success or fail parsing.
	 * @return An object (double value).
	 */
	public Object parseObject(String source, ParsePosition parsePosition)
	{
		return parse(source, parsePosition);
	}

	/**
	 * Parses source to find object representing scale.
	 *
	 * @param source String to parse.
	 * @param parsePosition Parse position which tracks success or fail parsing.
	 * @return A double value wrapped in class.
	 */
	public Double parse(String source, ParsePosition parsePosition)
	{
		if (parsePosition == null)
			throw new NullPointerException("Parse position is null");

		double scale = Double.NaN;

		int length = source.length();
		int startPosition = parsePosition.getIndex();
		int index = startPosition;

		// Pomi biae znaki
		while (index < length && Character.isWhitespace(source.charAt(index)))
			index++;
		parsePosition.setIndex(index);

		// Rozbierz licznik
		Number fieldObject = decimalFormat.parse(source, parsePosition);
		if (fieldObject == null)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		double parsedValue = fieldObject.doubleValue();

		// Sprawd czy liczba nie dodatnia
		if (parsedValue <= 0.0)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}

		index = parsePosition.getIndex();
		if (index == length)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}

		// Sprawd separator
		if (!source.regionMatches(index, delimeter, 0, delimeter.length()))
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		index += delimeter.length();
		parsePosition.setIndex(index);

		// Sprawd drug liczb
		fieldObject = decimalFormat.parse(source, parsePosition);
		if (fieldObject == null)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		scale = fieldObject.doubleValue();

		// Sprawd czy liczba jest dodatnia
		if (scale <= 0.0)
		{
			parsePosition.setIndex(startPosition);
			if (parsePosition.getErrorIndex() < startPosition)
				parsePosition.setErrorIndex(index);
			return null;
		}
		
		return new Double(parsedValue / scale);
	}

	/*
	 * -------------- Members ----------------
	 */

	/** Symbol. */
	private String delimeter = ":";

	/** Decimal formatter. */
	private DecimalFormat decimalFormat;
}
