/*
 * jET - An alternative scenery editor for the electic locomotive simulator, MaSzyna
 * Copyright (C) 2003  Adam Domaski
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package prv.doman.util;

import java.io.Reader;
import java.io.PushbackReader;
import java.io.StringReader;
import java.io.IOException;

/**
 * The stream tokenizer gets a stream and parses it into tokens.
 * <p>There are some predefined tokens:</p>
 * <ul>
 *   <li>numbers,
 *   <li>ends of lines,
 *   <li>ends of file,
 * </ul>
 * <p>and regular words.</p>
 * <p>Tokenizer may ignore slash slash comments (like in C++).</p>
 *
 * <h3>Using</h3>
 *
 * <p>This is an example of using this class.</p>
 * <blockquote>
 *   <pre>
 *     Tokenizer t = new Tokenizer(new FileReader("note.txt"));
 *     while (t.hasMoreTokens())
 *     {
 *        Token token = t.nextToken();
 *        ...
 *     }
 *   </pre>
 * </blockquote>
 *
 * <h3>Caution!</h3>
 * <p>This class is not optimized.</p>
 *
 * @author Adam Domaski
 * @version 0.1 30/08/2003
 * @see Token
 */
public class Tokenizer
{
	/**
	 * Creates tokenizer to tokenize reader input.
	 * After that tokenizer will not ignore cases and slash, slash comments.
	 * <p>It's much better to wrap input reader with buffering.</p>
	 *
	 * @param in Input to tokenize.
	 */
	public Tokenizer(Reader in)
	{
		this.in = new PushbackReader(in, 2);
	}

	/**
	 * Creates tokenizer to tokenize string.
	 *
	 * @param source String to tokenize.
	 */
	public Tokenizer(String source)
	{
		this(new StringReader(source));
	}

	/*
	 * ------------------ Tokens ---------------------
	 */

	/**
	 * Returns <code>true</code>, if in the input is at least one 
	 * token to read.
	 *
	 * @return <code>True</code>, if in the input is at least one
	 *   token to read.
	 * @exception IOException When some error occured.
	 */
	public boolean hasMoreTokens() throws IOException
	{
		int c;

		// Pomi biae znaki
		while ((c = read()) != -1 && ((c == ' ') || (c == '\t') || (c == '\f')))
			colCounter++;

		// Sprawd czy wystpi koniec pliku
		if (c == -1)
			return false;

		// Wyklucz tokeny z komentarza
		if (!ignoreSlash && (c == '/'))
		{
			c = read();
			if (c != '/')
			{
				unread('/');
				unread(c);
				return true;
			}
			else
			{
				colCounter++;

				// Pomi wszystko do znaku koca linii
				while ((c = read()) != -1 && ((c != '\n') && (c != '\r')))
					colCounter++;

				// Koniec pliku
				if (c == -1)
					return false;

				unread(c);
				return true;
			}
		}
		else
		{
			unread(c);
		}
		
		return true;
	}

	/**
	 * Gets next token from input and returns it.
	 *
	 * @return The next token.
	 * @exception IOException When an error occured.
	 */
	public Token nextToken() throws IOException
	{
		Token token;
		int c;

		// Pomi biae znaki
		while ((c = read()) != -1 && ((c == ' ') || (c == '\t') || (c == '\f')))
			colCounter++;

		// Sprawd czy osignito koniec pliku
		if (c == -1)
		{
			token = new Token();
			token.setWord("\n");
			token.setType(Token.TOKEN_EOF);
			token.setLine(lineCounter);
			token.setColumn(colCounter);
			return token;
		}

		// Sprawd czy osignito koniec linii
		if ((c == '\n') || (c == '\r'))
		{
			char[] cha;

			if (c == '\r')
			{
				c = read();
				if (c != '\n')
				{
					unread(c);
					cha = new char[1];
					cha[0] = '\r';
				}
				else
				{
					colCounter++;
					cha = new char[2];
					cha[0] = '\r';
					cha[1] = '\n';
				}
			}
			else
			{
				cha = new char[1];
				cha[0] = '\n';
			}

			// Stwrz token
			token = new Token();
			token.setWord(new String(cha));
			token.setType(Token.TOKEN_EOL);
			token.setColumn(colCounter);
			token.setLine(lineCounter);

			lineCounter++;
			colCounter = 0;

			return token;
		}

		unread(c);

		// Sprawd pozostao
		StringBuffer sb = new StringBuffer();
		int state = 1;
		boolean commentFlag = false;
		while (((c = read()) != -1) && (c != ' ') && (c != '\t') && (c != '\f')
			&& (c != '\n') && (c != '\r') && !commentFlag)
		{
			switch (state)
			{
				case 1:
					if (c == '-')
						state = 2;
					else if ((c >= '0') && (c <= '9'))
						state = 3;
					else
						state = 6;
					break;
				case 2:
					state = ((c >= '0') && (c <= '9')) ? 3 : 6;
					break;
				case 3:
					if ((c >= '0') && (c <= '9'))
						state = 3;
					else if (c == '.')
						state = 4;
					else
						state = 6;
					break;
				case 4:
					state = ((c >= '0') && (c <= '9')) ? 5 : 6;
					break;
				case 5:
					if (c >= '0' && c <= '9')
						state = 5;
					else if (c == 'E' || c == 'e')
						state = 7;
					else
						state = 6;
					break;
				case 7:
					if (c >= '0' && c <= '9')
						state = 9;
					else if (c == '-')
						state = 8;
					else
						state = 6;
					break;
				case 8:
					if (c >= '0' && c <= '9')
						state = 9;
					else
						state = 6;
					break;
				case 9:
					if (c >= '0' && c <= '9')
						state = 9;
					else
						state = 6;
					break;
			}

			// Sprawd czy token nie jest komentarzem
			if (!ignoreSlash && c == '/')
			{
				c = read();
				if (c == '/')
				{
					unread('/');
					unread('/');
					commentFlag = true;
				}
				else
				{
					unread(c);
					sb.append('/');
					colCounter++;
				}
			}
			else
			{
				sb.append((char) c);
				colCounter++;
			}

		}

		if (c != -1)
		{
			unread(c);
		}

		if (state == 5 || state == 3 || state == 9)
		{
			token = new NumberToken();
			token.setType(Token.TOKEN_NUMBER);
		}
		else
		{
			token = new Token();
			token.setType(Token.TOKEN_WORD);
		}
		token.setWord(sb.toString());
		token.setLine(lineCounter);
		token.setColumn(colCounter);

		return token;
	}

	/*
	 * -------------------- File ---------------------
	 */
	
	/**
	 * Returns current line number.
	 *
	 * @return The line number.
	 */
	public int getLine()
	{
		return lineCounter;
	}

	/**
	 * Reads the next character
	 */
	private int read() throws IOException
	{
		return in.read();
	}

	/**
	 * Unreads one character.
	 */
	private void unread(int ch) throws IOException
	{
		in.unread(ch);
	}

	/*
	 * ----------------- Conventions -----------------
	 */

	/**
	 * Sets slash slash comments ignoration.
	 * Slash slash comments begin with two slashes <code>//</code> and
	 * end at the end of line.
	 *
	 * @param b If <code>true</code>, slash slash comments will be ignored.
	 */
	public void setIgnoreSlashSlash(boolean b)
	{
		ignoreSlash = b;
	}

	/**
	 * Gets slash slash comments ingoration flag.
	 *
	 * @return <code>True</code> means that tokenizer ignores
	 *   slash slash comments.
	 */
	public boolean getIgnoreSlashSlash()
	{
		return ignoreSlash;
	}

	/*
	 * ------------------ Members --------------------
	 */
	
	/** Reader to tokenize. */
	private PushbackReader in = null;

	/** Line number counter. */
	private int lineCounter = 1;

	/** Column counter. */
	private int colCounter = 0;

	/** Ignore slash slash flag. */
	private boolean ignoreSlash = false;
}
